import argparse
import json
import os.path
import sys
import traceback

import appdirs

from .client import Batches, Groups, Permissions, Projects, Users
from .wrappers import BatchesWrapper, GroupsWrapper, PermissionsWrapper, ProjectsWrapper, \
    UsersWrapper
from .__version__ import __version__


config_choices = ['token', 'url', 'print']
config_help = """token  Set the API token
url    Set the base URL for the Turkle site. Ex: http://localhost:8000/
print  Print the current configuration values
"""

users_choices = ['list', 'create', 'retrieve', 'update']
users_help = """list      List all users as jsonl
create    Create new users
retrieve  Retrieve a user selected by a username or integer identifier
update    Update users
"""

groups_choices = ['list', 'create', 'retrieve', 'addusers']
groups_help = """list      List all groups as jsonl
create    Create new groups
retrieve  Retrieve a group selected by name or integer identifier
addusers  Add users to an existing group by passing their ids as list in file
"""

projects_choices = ['list', 'create', 'retrieve', 'update', 'batches']
projects_help = """list      List all projects as jsonl
create    Create new projects
retrieve  Retrieve a project based on integer identifier
update    Update projects
batches   List batches for a project
"""

batches_choices = ['list', 'create', 'retrieve', 'update', 'input', 'results', 'progress']
batches_help = """list      List all batches as jsonl
create    Create new batches
retrieve  Retrieve a batch based on integer identifier
update    Update batches
input     Download the input CSV
results   Download the current results CSV
progress  Get current progress information
"""

perm_choices = ['retrieve', 'add', 'replace']
perm_help = """retrieve  Retrieve permissions for a project or batch
add       Add users or groups to a project's or batch's permissions
replace   Replace a project's or batch's permissions
"""


class Cli:
    def __init__(self):
        self.debug = False

        self.parser = argparse.ArgumentParser(description='Turkle client help')
        self.parser.add_argument('-t', '--token', help='API token')
        self.parser.add_argument('-u', '--url', help='Base URL for the Turkle site')
        self.parser.add_argument('-d', '--debug', action='store_true',
                                 help='Detailed stack traces on errors')
        self.parser.add_argument('--version', action='store_true',
                                 help='Get the version of this client')
        self.update_title(self.parser, 'Object command')
        subparsers = self.parser.add_subparsers(dest='command')

        config_parser = subparsers.add_parser(
            'config',
            help='Set the token or url in the config.',
            formatter_class=argparse.RawTextHelpFormatter
        )
        self.update_title(config_parser, 'Parameter')
        config_parser.add_argument('subcommand', choices=config_choices, help=config_help)
        config_parser.add_argument('value', nargs='?', help='The value to set for the parameter')

        users_parser = subparsers.add_parser(
            'users',
            help='List, create, or update users.',
            formatter_class=argparse.RawTextHelpFormatter
        )
        self.update_title(users_parser)
        users_parser.add_argument('subcommand', choices=users_choices, help=users_help)
        users_parser.add_argument('--id', help='User id (integer) - for retrieve')
        users_parser.add_argument('--username', help='Username - for retrieve')
        users_parser.add_argument('--file', help='jsonl/json/csv file - required for create and update')

        groups_parser = subparsers.add_parser(
            'groups',
            help='List, create, update groups or add users to a group.',
            formatter_class=argparse.RawTextHelpFormatter
        )
        self.update_title(groups_parser)
        groups_parser.add_argument('subcommand', choices=groups_choices, help=groups_help)
        groups_parser.add_argument('--id', help='User id - required for retrieve')
        groups_parser.add_argument('--name', help='Group name - for retrieve')
        groups_parser.add_argument('--file', help='json/jsonl file - required for create or addusers')

        projects_parser = subparsers.add_parser(
            'projects',
            help='List, create, or update projects.',
            formatter_class=argparse.RawTextHelpFormatter
        )
        self.update_title(projects_parser)
        projects_parser.add_argument('subcommand', choices=projects_choices, help=projects_help)
        projects_parser.add_argument('--id', help='Project id - required for retrieve and batches')
        projects_parser.add_argument('--file', help='json/jsonl file - required for create or update')

        batches_parser = subparsers.add_parser(
            'batches',
            help='List, create, or update batches or get results and progress.',
            formatter_class=argparse.RawTextHelpFormatter
        )
        self.update_title(batches_parser)
        batches_parser.add_argument('subcommand', choices=batches_choices, help=batches_help)
        batches_parser.add_argument('--id', help='Batch id - required for retrieve')
        batches_parser.add_argument('--file', help='json/jsonl file - required for create or update')

        perm_parser = subparsers.add_parser(
            'permissions',
            help='List, add, or replace permissions.',
            formatter_class=argparse.RawTextHelpFormatter
        )
        self.update_title(perm_parser)
        perm_parser.add_argument('subcommand', choices=perm_choices, help=perm_help)
        perm_parser.add_argument('--pid', help='Project id')
        perm_parser.add_argument('--bid', help='Batch id')
        perm_parser.add_argument('--file', help='Jsonl file - required for add or replace')

    @staticmethod
    def update_title(parser, title='Subcommand'):
        parser._positionals.title = title

    def dispatch(self):
        args = self.parser.parse_args()
        self.debug = bool(args.debug)

        if args.version:
            print(f"Turkle client version {__version__}")
            return

        if not args.command:
            self.parser.print_help()
            return

        if args.command == 'config':
            if args.subcommand == 'print':
                config = self.load_config()
                if config:
                    print("Current configuration:")
                    for key, val in config.items():
                        print(f"  {key}: {val}")
                else:
                    print("No configuration found.")
            elif args.subcommand and args.value:
                self.set_config(args.subcommand, args.value)
                print(f"{args.subcommand} set to {args.value}")
            else:
                raise ValueError(f"{args.subcommand} not specified")
            return

        config = self.load_config()
        token = args.token or config.get('token')
        url = args.url or config.get('url')
        if not token:
            raise ValueError("API token not specified (use --token or config)")
        if not url:
            raise ValueError("API URL not specified (use --url or config)")

        # construct the class and method from the command and subcommand
        client = self.construct_client(args.command.capitalize(), url, token, self.debug)
        print(getattr(client, args.subcommand)(**vars(args)), end='')

    def construct_client(self, name, url, token, debug):
        # the wrapper handles interactions requiring multiple calls
        client_class = getattr(sys.modules[__name__], name)
        client = client_class(url, token, debug)
        wrapper_class = getattr(sys.modules[__name__], name + 'Wrapper')
        return wrapper_class(client)

    @property
    def config_dir(self):
        return appdirs.user_config_dir('turkle-client', 'HLTCOE')

    @property
    def config_file(self):
        return os.path.join(self.config_dir, 'config.json')

    def load_config(self):
        if os.path.exists(self.config_file):
            with open(self.config_file, 'r') as fh:
                config = json.load(fh)
                return config
        else:
            return {}

    def set_config(self, parameter, value):
        config = self.load_config()
        config[parameter] = value
        os.makedirs(self.config_dir, exist_ok=True)
        with open(self.config_file, 'w') as fh:
            json.dump(config, fh)


def main():
    cli = Cli()
    try:
        cli.dispatch()
    except Exception as e:
        if cli.debug:
            traceback.print_exc()
        else:
            print(f"Error: {e}")


if __name__ == '__main__':
    main()



import io
import json
import os

import requests

from .exceptions import TurkleClientException


class Client:
    """
    Base client for Turkle REST API

    The child classes are Users, Groups, Projects, Batches, and Permissions.
    Their methods return json/jsonl or csv data as a string.
    """
    def __init__(self, base_url, token, debug=False):
        """Construct a client

        Args:
            base_url (str): The URL of the Turkle site
            token (str): An authentication token for Turkle
        """
        self.base_url = base_url.rstrip('/')
        self.headers = {'Authorization': f'TOKEN {token}'}
        self.debug = debug

    class Urls:
        # child classes must set the list and detail url for that part of the API
        list = ""
        detail = ""

    def list(self):
        """List all instances (user, group, project, batch)

        Returns:
            str: jsonl where each line is an object
        """
        url = self.Urls.list.format(base=self.base_url)
        return self._walk(url)

    def retrieve(self, instance_id):
        """Retrieve an instance from an id (user, group, project, batch)

        Args:
            instance_id (int): Instance id

        Returns:
            str: instance encoded as json
        """
        url = self.Urls.detail.format(base=self.base_url, id=instance_id)
        response = self._get(url)
        return response.text

    def create(self, instance):
        """Create an instance (group, project, batch)

        Args:
            instance (dict): Instance fields as dict

        Returns:
            str: json representation of the created instance
        """
        if self.debug:
            print(f"Debug: Create object dict: {instance}")
        url = self.Urls.list.format(base=self.base_url)
        response = self._post(url, instance)
        return response.text

    def _walk(self, url,  **kwargs):
        jsonl = io.StringIO()
        data = {'next': url}
        while data['next']:
            response = self._get(data['next'], **kwargs)
            if response.status_code >= 400:
                self._handle_errors(response)
            data = response.json()
            for instance in data['results']:
                jsonl.write(json.dumps(instance, ensure_ascii=False) + os.linesep)
        return jsonl.getvalue()

    def _get(self, url, *args, **kwargs):
        try:
            response = requests.get(url, *args, **kwargs, headers=self.headers)
            if response.status_code >= 400:
                self._handle_errors(response)
            return response
        except requests.exceptions.ConnectionError:
            raise TurkleClientException(f"Unable to connect to {self.base_url}")

    def _post(self, url, data, *args, **kwargs):
        try:
            response = requests.post(url, *args, **kwargs, json=data, headers=self.headers)
            if response.status_code >= 400:
                self._handle_errors(response)
            return response
        except requests.exceptions.ConnectionError:
            raise TurkleClientException(f"Unable to connect to {self.base_url}")

    def _patch(self, url, data, *args, **kwargs):
        try:
            response = requests.patch(url, *args, **kwargs, json=data, headers=self.headers)
            if response.status_code >= 400:
                self._handle_errors(response)
            return response
        except requests.exceptions.ConnectionError:
            raise TurkleClientException(f"Unable to connect to {self.base_url}")

    def _put(self, url, data, *args, **kwargs):
        try:
            response = requests.put(url, *args, **kwargs, json=data, headers=self.headers)
            if response.status_code >= 400:
                self._handle_errors(response)
            return response
        except requests.exceptions.ConnectionError:
            raise TurkleClientException(f"Unable to connect to {self.base_url}")

    def _handle_errors(self, response):
        data = response.json()
        if data:
            if 'detail' in data:
                raise TurkleClientException(data['detail'])
            else:
                # grab the first error
                parts = next(iter(data.items()))
                raise TurkleClientException(f"{parts[0]} - {parts[1][0]}")


class Users(Client):
    class Urls:
        list = "{base}/api/users/"
        detail = "{base}/api/users/{id}/"
        username = "{base}/api/users/username/{username}/"

    def retrieve_by_username(self, username):
        """Retrieve a user from a username

        Args:
            username (str): Username
        Returns:
            str: user object as json
        """
        url = self.Urls.username.format(base=self.base_url, username=username)
        response = self._get(url)
        return response.text

    def create(self, user):
        """Create a user

        Args:
            user (dict): User fields as dict

        Returns:
            str: json representation of the created user
        """
        if self.debug:
            print(f"Debug: New user dict: {user}")
        url = self.Urls.list.format(base=self.base_url)
        response = self._post(url, user)
        return response.text

    def update(self, user):
        """Update a user

        Args:
            user (dict): User fields as dict including id

        Returns:
            str: json representation of the updated user
        """
        if self.debug:
            print(f"Debug: Updated user dict: {user}")
        url = self.Urls.detail.format(base=self.base_url, id=user['id'])
        response = self._patch(url, user)
        return response.text


class Groups(Client):
    class Urls:
        list = "{base}/api/groups/"
        detail = "{base}/api/groups/{id}/"
        name = "{base}/api/groups/name/{name}/"
        addusers = "{base}/api/groups/{id}/users/"

    def retrieve_by_name(self, name):
        """Retrieve groups from a name

        Args:
            name (str): Group name

        Returns:
            str: jsonl with each line being a group that has that name
        """
        url = self.Urls.name.format(base=self.base_url, name=name)
        return self._walk(url)

    def addusers(self, group_id, user_ids, **kwargs):
        """Add users to a group

        Args:
            group_id (int): Group id
            user_ids (list): List of User ids

        Returns:
            str: json of the updated group
        """
        url = self.Urls.addusers.format(base=self.base_url, id=group_id)
        data = {'users': user_ids}
        response = self._post(url, data)
        return response.text


class Projects(Client):
    class Urls:
        list = "{base}/api/projects/"
        detail = "{base}/api/projects/{id}/"
        batches = "{base}/api/projects/{id}/batches/"

    def create(self, project):
        """Create a project

        Args:
            project (dict): Project fields as a dict

        Returns:
            str: json representation of the created project
        """
        url = self.Urls.list.format(base=self.base_url)
        response = self._post(url, project)
        return response.text

    def update(self, project):
        """Update a project

        Args:
            project (dict): Project fields including the id

        Returns:
            str: json representation of the updated project
        """
        url = self.Urls.detail.format(base=self.base_url, id=project['id'])
        response = self._patch(url, project)
        return response.text

    def batches(self, project_id):
        """List all batches for a project

        Args:
            project_id (int): Project id

        Returns:
            str: jsonl where each line is a batch object
        """
        url = self.Urls.batches.format(base=self.base_url, id=project_id)
        return self._walk(url)


class Batches(Client):
    class Urls:
        list = "{base}/api/batches/"
        detail = "{base}/api/batches/{id}/"
        input = "{base}/api/batches/{id}/input/"
        results = "{base}/api/batches/{id}/results/"
        progress = "{base}/api/batches/{id}/progress/"

    def create(self, batch):
        """Create a batch

        Args:
            batch (dict): Batch fields as a dict

        Returns:
            str: json representation of the created batch
        """
        url = self.Urls.list.format(base=self.base_url)
        response = self._post(url, batch)
        return response.text

    def update(self, batch):
        """Update a batch

        Cannot update the CSV data. See addtasks to add additional tasks.

        Args:
            batch (dict): Batch fields as a dict including the id

        Returns:
            str: json representations of the updated batch
        """
        url = self.Urls.detail.format(base=self.base_url, id=batch['id'])
        response = self._patch(url, batch)
        return response.text

    def input(self, batch_id):
        """Get the input CSV for the batch

        Args:
            batch_id (int): Batch id

        Returns:
             str: CSV data as a string
        """
        url = self.Urls.input.format(base=self.base_url, id=batch_id)
        response = self._get(url)
        return response.text

    def results(self, batch_id):
        """Get the results CSV for the batch

        Args:
            batch_id (int): Batch id

        Returns:
             str: CSV data as a string
        """
        url = self.Urls.results.format(base=self.base_url, id=batch_id)
        response = self._get(url)
        return response.text

    def progress(self, batch_id):
        """Get the progress information for the batch

        Args:
            batch_id (int): batch id

        Returns:
             str: json progress object
        """
        url = self.Urls.progress.format(base=self.base_url, id=batch_id)
        response = self._get(url)
        return response.text


class Permissions(Client):
    class Urls:
        projects = "{base}/api/projects/{id}/permissions/"
        batches = "{base}/api/batches/{id}/permissions/"

    PROJECT = 'project'
    BATCH = 'batch'

    def _get_url(self, instance_type, instance_id):
        if instance_type == self.PROJECT:
            url = self.Urls.projects.format(base=self.base_url, id=instance_id)
        elif instance_type == self.BATCH:
            url = self.Urls.batches.format(base=self.base_url, id=instance_id)
        else:
            raise TurkleClientException(f"Unrecognized instance type: {instance_type}")
        return url

    def get(self, instance_type, instance_id):
        """Get the permissions for the project or batch

        Args:
            instance_type (str): Name of the type (project, batch)
            instance_id (int): Id of the project or batch

        Returns:
            str: json representation of the permissions
        """
        url = self._get_url(instance_type, instance_id)
        response = self._get(url)
        return response.text

    def add(self, instance_type, instance_id, permissions):
        """Add additional users and groups to the permissions

        Args:
            instance_type (str): Name of the type (project, batch)
            instance_id (int): Id of the project or batch
            permissions (dict): Dictionary with keys 'users' and 'groups' for lists of ids

        Returns:
            str: json representation of the updated permissions
        """
        url = self._get_url(instance_type, instance_id)
        response = self._post(url, permissions)
        return response.text

    def replace(self, instance_type, instance_id, permissions):
        """Replace the permissions

        Args:
            instance_type (str): Name of the type (project, batch)
            instance_id (int): Id of the project or batch
            permissions (dict): Dictionary with keys 'users' and 'groups' for lists of ids

        Returns:
            str: json representation of the updated permissions
        """
        url = self._get_url(instance_type, instance_id)
        response = self._put(url, permissions)
        return response.text

    def list(self):
        raise NotImplementedError()

    def create(self, instance):
        raise NotImplementedError()

    def retrieve(self, instance_id):
        raise NotImplementedError()



import csv
import json
import os.path

from .client import Permissions
from .exceptions import TurkleClientException


def plural(num, single, mult):
    return f"{num} {single if num == 1 else mult}"


def load_records(file_path, exts=None):
    """
    Yields dictionaries from a .jsonl, .json or .csv file.

    Args:
        file_path (str): Path to the input file
        exts (list): List of extensions to support

    Returns:
        Iterator: iterator over dictionaries
    """
    exts = exts if exts else ['.jsonl', '.json', '.csv']
    ext = os.path.splitext(file_path)[1].lower()
    if ext not in exts:
        raise ValueError(f"Unsupported file format: {ext}")

    try:
        with open(file_path, 'r', encoding='utf-8') as fh:
            if ext == '.jsonl':
                for lineno, line in enumerate(fh, start=1):
                    if not line.strip():
                        continue  # skip blank lines
                    try:
                        yield json.loads(line)
                    except json.JSONDecodeError as e:
                        raise ValueError(f"Invalid JSON on line {lineno} in {file_path}: {e}")
            elif ext == '.csv':
                reader = csv.DictReader(fh)
                for row in reader:
                    yield row
            elif ext == '.json':
                yield json.load(fh)
            else:
                raise ValueError(f"Unsupported file format: {ext}")
    except OSError as e:
        raise ValueError(f"Could not open file {file_path}: {e}")


class Wrapper:
    """
    Client wrappers that massage input and output to match expectations for the CLI
    """
    def __init__(self, client):
        self.client = client

    def list(self, **kwargs):
        return self.client.list()


class UsersWrapper(Wrapper):
    def retrieve(self, id, username, **kwargs):
        if id:
            return self.client.retrieve(id) + "\n"
        elif username:
            return self.client.retrieve_by_username(username) + "\n"
        else:
            raise TurkleClientException("--id or --username must be set for 'users retrieve'")

    def create(self, file, **kwargs):
        if not file:
            raise TurkleClientException("--file must be set for 'users create'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file), start=1):
                self.client.create(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on line {lineno} in {file}: {e}")

        return f"{plural(lineno, 'user', 'users')} created\n"

    def update(self, file, **kwargs):
        if not file:
            raise TurkleClientException("--file must be set for 'users update'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file), start=1):
                self.client.update(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on line {lineno} in {file}: {e}")

        return f"{plural(lineno, 'user', 'users')} updated\n"


class GroupsWrapper(Wrapper):
    def retrieve(self, id, name, **kwargs):
        if id:
            return self.client.retrieve(id) + "\n"
        elif name:
            return self.client.retrieve_by_name(name)
        else:
            raise TurkleClientException("--id or --name must be set for 'groups retrieve'")

    def create(self, file, **kwargs):
        if not file:
            raise ValueError("--file must be set for 'groups create'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file, [".jsonl", ".json"]), start=1):
                self.client.create(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on object {lineno} in {file}: {e}")

        return f"{plural(lineno, 'group', 'groups')} created\n"

    def addusers(self, id, file, **kwargs):
        # file contains json encoded list of user ids
        if not id:
            raise ValueError("--id must be set for 'groups addusers'")
        if not file:
            raise ValueError("--file must be set for 'groups addusers'")

        try:
            user_ids = next(load_records(file, [".json"]))
            self.client.addusers(id, user_ids)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure in {file}: {e}")
        return f"{plural(len(user_ids), 'user', 'users')} added to the group\n"


class ProjectsWrapper(Wrapper):
    def retrieve(self, id, **kwargs):
        if not id:
            raise TurkleClientException("--id must be set for 'projects retrieve'")
        return self.client.retrieve(id) + "\n"

    def create(self, file, **kwargs):
        if not file:
            raise TurkleClientException("--file must be set for 'projects create'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file, [".jsonl", ".json"]), start=1):
                if 'html_template' not in obj:
                    with open(os.path.expanduser(obj['filename']), 'r') as template_fh:
                        obj['html_template'] = template_fh.read()
                        obj['filename'] = os.path.basename(obj['filename'])
                self.client.create(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on object {lineno} in {file}: {e}")

        return f"{plural(lineno, 'project', 'projects')} created\n"

    def update(self, file, **kwargs):
        if not file:
            raise TurkleClientException("--file must be set for 'projects update'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file, [".jsonl", ".json"]), start=1):
                if 'html_template' not in obj:
                    with open(os.path.expanduser(obj['filename']), 'r') as template_fh:
                        obj['html_template'] = template_fh.read()
                        obj['filename'] = os.path.basename(obj['filename'])
                self.client.create(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on object {lineno} in {file}: {e}")

        return f"{plural(lineno, 'project', 'projects')} updated\n"

    def batches(self, id, **kwargs):
        if not id:
            raise TurkleClientException("--id must be set for 'projects batches'")
        return self.client.batches(id)


class BatchesWrapper(Wrapper):
    def retrieve(self, id, **kwargs):
        if not id:
            raise TurkleClientException("--id must be set for 'batches retrieve'")
        return self.client.retrieve(id) + "\n"

    def create(self, file, **kwargs):
        if not file:
            raise TurkleClientException("--file must be set for 'batches create'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file, [".jsonl", ".json"]), start=1):
                with open(os.path.expanduser(obj['filename']), 'r') as csv_fh:
                    obj['csv_text'] = csv_fh.read()
                    obj['filename'] = os.path.basename(obj['filename'])
                self.client.create(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on object {lineno} in {file}: {e}")

        return f"{plural(lineno, 'batch', 'batches')} created\n"

    def update(self, file, **kwargs):
        if not file:
            raise TurkleClientException("--file must be set for 'batches update'")

        lineno = 0
        try:
            for lineno, obj in enumerate(load_records(file, [".jsonl", ".json"]), start=1):
                if 'filename' in obj:
                    with open(os.path.expanduser(obj['filename']), 'r') as csv_fh:
                        obj['csv_text'] = csv_fh.read()
                        obj['filename'] = os.path.basename(obj['filename'])
                self.client.update(obj)
        except TurkleClientException as e:
            raise TurkleClientException(f"Failure on object {lineno} in {file}: {e}")

        return f"{plural(lineno, 'batch', 'batches')} updated\n"

    def input(self, id, **kwargs):
        if not id:
            raise TurkleClientException("--id must be set for 'batches input'")
        return self.client.input(id)

    def progress(self, id, **kwargs):
        if not id:
            raise TurkleClientException("--id must be set for 'batches progress'")
        return self.client.progress(id) + "\n"

    def results(self, id, **kwargs):
        if not id:
            raise TurkleClientException("--id must be set for 'batches results'")
        return self.client.results(id)


class PermissionsWrapper(Wrapper):
    def _prepare_args(self, pid, bid):
        if pid:
            return Permissions.PROJECT, pid
        else:
            return Permissions.BATCH, bid

    def retrieve(self, pid, bid, **kwargs):
        if not pid and not bid:
            raise TurkleClientException("--pid or --bid is required for 'permissions retrieve'")
        text = self.client.get(*self._prepare_args(pid, bid))
        return text + "\n"

    def add(self, pid, bid, file, **kwargs):
        if not pid and not bid:
            raise TurkleClientException("--pid or --bid is required for 'permissions add'")
        if not file:
            raise ValueError("--file must be set for 'permissions add'")
        with open(file, 'r') as fh:
            data = json.load(fh)
            text = self.client.add(*self._prepare_args(pid, bid), data)
            return text + "\n"

    def replace(self, pid, bid, file, **kwargs):
        if not pid and not bid:
            raise TurkleClientException("--pid or --bid is required for 'permissions replace'")
        if not file:
            raise ValueError("--file must be set for 'permissions replace'")
        with open(file, 'r') as fh:
            data = json.load(fh)
            text = self.client.replace(*self._prepare_args(pid, bid), data)
            return text + "\n"



import pytest
import vcr

from turkle_client.client import Client
from turkle_client.exceptions import TurkleClientException

my_vcr = vcr.VCR(
    cassette_library_dir='tests/fixtures/cassettes/client/',
)


@my_vcr.use_cassette()
def test_bad_token():
    client = Client("http://localhost:8000/", "bad_token")
    with pytest.raises(TurkleClientException, match="Invalid token"):
        client._get("http://localhost:8000/api/users/")


@my_vcr.use_cassette()
def test_404():
    client = Client("http://localhost:8000/", "41dcbb22264dd60c5232383fc844dbbab4839146")
    with pytest.raises(TurkleClientException, match="Not found"):
        client._get("http://localhost:8000/api/users/999999/")



import pytest
import vcr

from turkle_client.client import Users

my_vcr = vcr.VCR(
    cassette_library_dir='tests/fixtures/cassettes/users/',
)


@my_vcr.use_cassette()
def test_retrieve():
    client = Users("http://localhost:8000/", "41dcbb22264dd60c5232383fc844dbbab4839146")
    text = client.retrieve(1)
    assert 'AnonymousUser' in text



from typing import Optional

import pydantic


class User(pydantic.BaseModel):
    username: str
    password: str
    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]
    is_active: Optional[bool]
    is_staff: Optional[bool]
    is_superuser: Optional[bool]
